@@ -7,15 +7,45 @@ import datetime
 import uuid
 import binascii
 import time
-from urllib.parse import urlparse, parse_qs
+import codecs
+from urllib.parse import urlparse, parse_qs, quote_plus
 import os
-from cryptography.hazmat.primitives import hashes
+from cryptography.hazmat.primitives import serialization, padding, hashes
 from cryptography.hazmat.primitives.kdf.kbkdf import CounterLocation, KBKDFHMAC, Mode
 from cryptography.hazmat.backends import default_backend
+from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
 import requests
 import adal
 import jwt
 
+
+
+WELLKNOWN_RESOURCES = {
+    "msgraph": "https://graph.microsoft.com/",
+    "aadgraph": "https://graph.windows.net/",
+    "devicereg": "urn:ms-drs:enterpriseregistration.windows.net",
+    "drs": "urn:ms-drs:enterpriseregistration.windows.net",
+    "azrm": "https://management.core.windows.net/",
+    "azurerm": "https://management.core.windows.net/",
+}
+
+WELLKNOWN_CLIENTS = {
+    "aadps": "1b730954-1685-4b74-9bfd-dac224a7b894",
+    "azcli": "04b07795-8ddb-461a-bbee-02f9e1bf7b46",
+    "teams": "1fec8e78-bce4-4aaf-ab1b-5451cc387264",
+    "msteams": "1fec8e78-bce4-4aaf-ab1b-5451cc387264",
+    "azps": "1950a258-227b-4e31-a9cf-717495945fc2",
+}
+
+def get_data(data):
+    return base64.urlsafe_b64decode(data+('='*(len(data)%4)))
+
+class AuthenticationException(Exception):
+    """
+    Generic exception we can throw when auth fails so that the error
+    goes back to the user.
+    """
+
 class Authentication():
     """
     Authentication class for ROADtools
@@ -24,7 +54,8 @@ class Authentication():
         self.username = username
         self.password = password
         self.tenant = tenant
-        self.client_id = client_id
+        self.client_id = None
+        self.set_client_id(client_id)
         self.resource_uri = 'https://graph.windows.net/'
         self.tokendata = {}
         self.refresh_token = None
@@ -43,6 +74,18 @@ class Authentication():
             return 'https://login.microsoftonline.com/{}'.format(self.tenant)
         return 'https://login.microsoftonline.com/common'
 
+    def set_client_id(self, clid):
+        """
+        Sets client ID to use (accepts aliases)
+        """
+        self.client_id = self.lookup_client_id(clid)
+
+    def set_resource_uri(self, uri):
+        """
+        Sets resource URI to use (accepts aliases)
+        """
+        self.resource_uri = self.lookup_resource_uri(uri)
+
     def authenticate_device_code(self):
         """
         Authenticate the end-user using device auth.
@@ -67,7 +110,6 @@ class Authentication():
 
         return self.tokendata
 
-
     def authenticate_as_app(self):
         """
         Authenticate with an APP id + secret (password credentials assigned to serviceprinicpal)
@@ -107,6 +149,84 @@ class Authentication():
         self.tokendata['tenantId'] = inputdata['tid']
         return self.tokendata
 
+    def authenticate_with_code_native(self, code, redirurl, client_secret=None, pkce_secret=None):
+        """
+        Authenticate with a code plus optional secret in case of a non-public app (authorization grant)
+        Native ROADlib implementation without adal requirement - also supports PKCE
+        """
+        authority_uri = self.get_authority_url()
+        data = {
+            "client_id": self.client_id,
+            "grant_type": "authorization_code",
+            "code": code,
+            "redirect_uri": redirurl,
+            "resource": self.resource_uri,
+        }
+        if client_secret:
+            data['client_secret'] = client_secret
+        if pkce_secret:
+            raise NotImplementedError
+        res = requests.post(f"{authority_uri}/oauth2/token", data=data)
+        if res.status_code != 200:
+            raise AuthenticationException(res.text)
+        tokenreply = res.json()
+        self.tokendata = self.tokenreply_to_tokendata(tokenreply)
+        return self.tokendata
+
+    def authenticate_with_code_encrypted(self, code, sessionkey, redirurl):
+        '''
+        Encrypted code redemption. Like normal code flow but requires
+        session key to decrypt response.
+        '''
+        authority_uri = self.get_authority_url()
+        data = {
+            "grant_type": "authorization_code",
+            "code": code,
+            "redirect_uri": redirurl,
+            "client_id": self.client_id,
+            "client_info":1,
+            "windows_api_version":"2.0"
+        }
+        res = requests.post(f"{authority_uri}/oauth2/token", data=data)
+        if res.status_code != 200:
+            raise AuthenticationException(res.text)
+        prtdata = res.text
+        data = self.decrypt_auth_response(prtdata, sessionkey, asjson=True)
+        return data
+
+    def build_auth_url(self, redirurl, response_type, scope=None, state=None):
+        '''
+        Build authorize URL. Can be v2 by specifying scope, otherwise defaults
+        to v1 with resource
+        '''
+        urlt_v2 = 'https://login.microsoftonline.com/{3}/oauth2/v2.0/authorize?response_type={4}&client_id={0}&scope={2}&redirect_uri={1}&state={5}'
+        urlt_v1 = 'https://login.microsoftonline.com/{3}/oauth2/authorize?response_type={4}&client_id={0}&resource={2}&redirect_uri={1}&state={5}'
+        if not state:
+            state = str(uuid.uuid4())
+        if not self.tenant:
+            tenant = 'common'
+        else:
+            tenant = self.tenant
+        if scope:
+            # v2
+            return urlt_v2.format(
+                quote_plus(self.client_id),
+                quote_plus(redirurl),
+                quote_plus(scope),
+                quote_plus(tenant),
+                quote_plus(response_type),
+                quote_plus(state)
+            )
+        # Else default to v1 identity endpoint
+        return urlt_v1.format(
+            quote_plus(self.client_id),
+            quote_plus(redirurl),
+            quote_plus(self.resource_uri),
+            quote_plus(tenant),
+            quote_plus(response_type),
+            quote_plus(state)
+        )
+
     def create_prt_cookie_kdf_ver_2(self, prt, sessionkey, nonce=None):
         """
         KDF version 2 cookie construction
@@ -116,15 +236,20 @@ class Authentication():
             'ctx': base64.b64encode(context).decode('utf-8'), #.rstrip('=')
             'kdf_ver': 2
         }
-        if not nonce:
-            nonce = self.get_prt_cookie_nonce()
-            if not nonce:
-                return False
-        payload = {
-            "refresh_token": prt,
-            "is_primary": "true",
-            "request_nonce": nonce
-        }
+        # Nonce should be requested by calling function, otherwise
+        # old time based model is used
+        if nonce:
+            payload = {
+                "refresh_token": prt,
+                "is_primary": "true",
+                "request_nonce": nonce
+            }
+        else:
+            payload = {
+                "refresh_token": prt,
+                "is_primary": "true",
+                "iat": '{}'.format(int(time.time()))
+            }
         # Sign with random key just to get jwt body in right encoding
         tempjwt = jwt.encode(payload, os.urandom(32), algorithm='HS256', headers=headers)
         jbody = tempjwt.split('.')[1]
@@ -203,14 +328,47 @@ class Authentication():
         derived_key = kdf.derive(sessionkey)
         return context, derived_key
 
+    def decrypt_auth_response(self, responsedata, sessionkey, asjson=False):
+        """
+        Decrypt an encrypted authentication response, which is a JWE
+        encrypted using the sessionkey
+        """
+        if responsedata[:2] == '{"':
+            # This doesn't appear encrypted
+            if asjson:
+                return json.loads(responsedata)
+            return responsedata
+        dataparts = responsedata.split('.')
+
+        headers = json.loads(get_data(dataparts[0]))
+        _, derived_key = self.calculate_derived_key(sessionkey, base64.b64decode(headers['ctx']))
+        data = dataparts[3]
+        iv = dataparts[2]
+
+        cipher = Cipher(algorithms.AES(derived_key), modes.CBC(get_data(iv)))
+        decryptor = cipher.decryptor()
+        decrypted_data = decryptor.update(get_data(data)) + decryptor.finalize()
+        unpadder = padding.PKCS7(128).unpadder()
+        depadded_data = unpadder.update(decrypted_data) + unpadder.finalize()
+        if asjson:
+            jdata = json.loads(depadded_data)
+            return jdata
+        else:
+            return depadded_data
+
     def get_srv_challenge(self):
+        """
+        Request server challenge (nonce) to use with a PRT
+        """
         data = {'grant_type':'srv_challenge'}
         res = requests.post('https://login.microsoftonline.com/common/oauth2/token', data=data)
         return res.json()
 
     def get_prt_cookie_nonce(self):
         """
-        Request a nonce to sign in with
+        Request a nonce to sign in with. This nonce is taken from the sign-in page, which
+        is how Chrome processes it, but it could probably also be obtained using the much
+        simpler request from the get_srv_challenge function.
         """
         ses = requests.session()
         params = {
@@ -443,7 +601,7 @@ class Authentication():
             return None
         secret = derived_key.replace(' ','')
         sdata = binascii.unhexlify(secret)
-        return secret
+        return sdata
 
     @staticmethod
     def ensure_binary_sessionkey(sessionkey):
@@ -471,16 +629,107 @@ class Authentication():
             prt = base64.b64decode(prt+('='*(len(prt)%4))).decode('utf-8')
         return prt
 
+    @staticmethod
+    def parse_accesstoken(token):
+        tokenparts = token.split('.')
+        tokendata = json.loads(get_data(tokenparts[1]))
+        tokenobject = {
+            'accessToken': token,
+            'tokenType': 'Bearer',
+            'expiresOn': datetime.datetime.fromtimestamp(tokendata['exp']).strftime('%Y-%m-%d %H:%M:%S'),
+            'tenantId': tokendata['tid'],
+            '_clientId': tokendata['appid']
+        }
+        return tokenobject, tokendata
+
+    @staticmethod
+    def tokenreply_to_tokendata(tokenreply):
+        """
+        Convert /token reply from Azure to ADAL compatible object
+        """
+        tokenobject = {
+            'tokenType': tokenreply['token_type'],
+            'expiresOn': datetime.datetime.fromtimestamp(int(tokenreply['expires_on'])).strftime('%Y-%m-%d %H:%M:%S')
+        }
+        translate_map = {
+            'access_token': 'accessToken',
+            'refresh_token': 'refreshToken',
+            'id_token': 'idToken',
+            'token_type': 'tokenType'
+        }
+        for newname, oldname in translate_map.items():
+            if newname in tokenreply:
+                tokenobject[oldname] = tokenreply[newname]
+        return tokenobject
+
+    @staticmethod
+    def parse_compact_jwe(jwe, verbose=False, decode_header=True):
+        """
+        Parse compact JWE according to
+        https://datatracker.ietf.org/doc/html/rfc7516#section-3.1
+        """
+        dataparts = jwe.split('.')
+        header, enc_key, iv, ciphertext, auth_tag = dataparts
+        parsed_header = json.loads(get_data(header))
+        if verbose:
+            print("Header (decoded):")
+            print(json.dumps(parsed_header, sort_keys=True, indent=4))
+            print("Encrypted key:")
+            print(enc_key)
+            print('IV:')
+            print(iv)
+            print("Ciphertext:")
+            print(ciphertext)
+            print("Auth tag:")
+            print(auth_tag)
+        if decode_header:
+            return parsed_header, enc_key, iv, ciphertext, auth_tag
+        return header, enc_key, iv, ciphertext, auth_tag
+
+    @staticmethod
+    def parse_jwt(jwt):
+        """
+        Simple JWT parsing function
+        returns header and body as dict and signature as bytes
+        """
+        dataparts = jwt.split('.')
+        header = json.loads(get_data(dataparts[0]))
+        body = json.loads(get_data(dataparts[1]))
+        signature = get_data(dataparts[2])
+        return header, body, signature
+
+    @staticmethod
+    def lookup_resource_uri(uri):
+        """
+        Translate resource URI aliases
+        """
+        try:
+            resolved = WELLKNOWN_RESOURCES[uri.lower()]
+            return resolved
+        except KeyError:
+            return uri
+
+    @staticmethod
+    def lookup_client_id(clid):
+        """
+        Translate client ID aliases
+        """
+        try:
+            resolved = WELLKNOWN_CLIENTS[clid.lower()]
+            return resolved
+        except KeyError:
+            return clid
+
     def parse_args(self, args):
         self.username = args.username
         self.password = args.password
         self.tenant = args.tenant
-        self.client_id = args.client
+        self.set_client_id(args.client)
         self.access_token = args.access_token
         self.refresh_token = args.refresh_token
         self.outfile = args.tokenfile
         self.debug = args.debug
-        self.resource_uri = args.resource
+        self.set_resource_uri(args.resource)
 
         if not self.username is None and self.password is None:
             self.password = getpass.getpass()
@@ -489,18 +738,14 @@ class Authentication():
         if self.tokendata:
             return self.tokendata
         if self.refresh_token and not self.access_token:
-            token_data = {'refreshToken': self.refresh_token}
+            if self.refresh_token == 'file':
+                with codecs.open(args.tokenfile, 'r', 'utf-8') as infile:
+                    token_data = json.load(infile)
+            else:
+                token_data = {'refreshToken': self.refresh_token}
             return self.authenticate_with_refresh(token_data)
         if self.access_token and not self.refresh_token:
-            tokens = self.access_token.split('.')
-            inputdata = json.loads(base64.b64decode(tokens[1]+('='*(len(tokens[1])%4))))
-            self.tokendata = {
-                'accessToken': self.access_token,
-                'tokenType': 'Bearer',
-                'expiresOn': datetime.datetime.fromtimestamp(inputdata['exp']).strftime('%Y-%m-%d %H:%M:%S'),
-                'tenantId': inputdata['tid'],
-                '_clientId': inputdata['appid']
-            }
+            self.tokendata, _ = self.parse_accesstoken(self.access_token)
             return self.tokendata
         if self.username and self.password:
             return self.authenticate_username_password()
@@ -538,7 +783,7 @@ class Authentication():
         if args.tokens_stdout:
             sys.stdout.write(json.dumps(self.tokendata))
         else:
-            with open(self.outfile, 'w') as outfile:
+            with codecs.open(self.outfile, 'w', 'utf-8') as outfile:
                 json.dump(self.tokendata, outfile)
             print('Tokens were written to {}'.format(self.outfile))
 
